;; Talking Scale.
;;
;; Copyright 2020 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; CPU: MSP430G2231
;; SparkFun HX711 Load Sensor for the scale.
;; SparkFun QWIIC MP3 Trigger using i2c for sound.

.include "msp430x2xx.inc"

RAM equ 0x0200
DEVICE_ID equ RAM
I2C_ACK equ RAM+1
DEBUG_0 equ RAM+2
DEBUG_1 equ RAM+3
REMAINDER equ RAM+4
SCALE_VALUE_0 equ RAM+6
SCALE_VALUE equ RAM+16
SCALE_DIFF equ RAM+20
SCALE_DECIMAL equ RAM+22
SCALE_FRACTIONAL equ RAM+24
SCALE_LAST equ RAM+26

SCALE_DATA equ 0x01
SCALE_CLOCK equ 0x02
LED_0 equ 0x04
MP3_RST equ 0x10
MP3_INT equ 0x20
MP3_SCL equ 0x40
MP3_SDA equ 0x80

;  r4 = Interrupt counter
;  r5 =
;  r6 =
;  r7 =
;  r8 =
;  r9 =
; r10 =
; r11 =
; r12 = pointer to scale value
; r13 = temp in main
; r14 = function param
; r15 = function param and return value

.org 0xf800
start:
  ;; Turn off watchdog
  mov.w #(WDTPW|WDTHOLD), &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0280, SP

  ;; Set MCLK to 1.4 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_8, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set SMCLK to 32.768kHz external crystal
  ;mov.b #XCAP_3, &BCSCTL3

  ;; Setup output pins
  ;; P1.0 = Load Cell Data
  ;; P1.1 = Load Cell Clock
  ;; P1.2 = led 0
  ;; P1.4 = /RST
  ;; P1.5 = /INT
  ;; P1.6 = i2c SCL
  ;; P1.7 = i2c SDA
  mov.b #MP3_RST|LED_0|SCALE_CLOCK, &P1DIR
  mov.b #0, &P1OUT
  mov.b #0, &P1SEL
  mov.b #0, &P1SEL2

  ;; Setup i2c
  mov.b #USISWRST, &USICTL0
  bis.b #USIPE7|USIPE6|USIMST, &USICTL0
  mov.b #USII2C, &USICTL1
  mov.b #USIDIV_3|USISSEL_2|USICKPL, &USICKCTL
  bis.b #USISCLREL, &USICNT      ; let the clock float (pulled high by res)
  bic.b #USISWRST, &USICTL0      ; clear reset

  ;; Setup Timer A
  ;mov.w #32768, &TACCR0
  ;mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT to TACCR0
  ;mov.w #CCIE, &TACCTL0
  ;mov.w #0, &TACCTL1

  ;; Turn on interrupts
  eint

  ;mov.w #0x0000, &SCALE_VALUE+0
  ;mov.w #0x0000, &SCALE_VALUE+2

  ;; To help debugging, clear out some RAM.
  mov.w #0x200, r15
memset:
  mov.w #0, 0(r15)
  add.w #2, r15
  cmp.w #0x230, r15
  jnz memset

  ;; Take mp3 player out of reset
  bis.b #MP3_RST, &P1OUT

  ;; Wait up to 3 seconds for the card MP3 board to init (really
  ;; only needs 1.5 seconds, but just being safe).
;  mov.w #0, r4
;mp3_board_init_wait:
;  cmp.w #3, r4
;  jnz mp3_board_init_wait
  call #delay_3s

  call #get_id

  mov.b #22, r14
  call #set_volume

.if 0
  ;; DEBUG
  call #is_playing
  mov.b r15, DEBUG_0
  ;; DEBUG

  mov.b #1, r14
  call #play_file
  call #delay_50ms

  ;; DEBUG
  call #is_playing
  mov.b r15, DEBUG_1
  ;; DEBUG
.endif

  call #wait_for_mp3_player

  ;; Get a reading from the scale with no weight on it so when
  ;; there is a weight this value can be subtracted.
  call #read_scale
  mov.w &SCALE_VALUE, r15
  swpb r15
  mov.w r15, &SCALE_VALUE_0

main:
  call #read_scale
  mov.w &SCALE_VALUE, r15
  swpb r15
  cmp.w #0xff00, r15
  jhs main
  call #speak_weight
  jmp main

read_scale:
  bic.b #LED_0, &P1OUT
  bit.b #0x01, &P1IN
  jnz read_scale
  mov.w #SCALE_VALUE, r12
  mov.b #3, r13
read_scale_next_byte:
  mov.b #0, r15
  mov.b #8, r14
read_scale_next_bit:
  bis.b #SCALE_CLOCK, &P1OUT
  bic.b #SCALE_CLOCK, &P1OUT
  rla.b r15
  bit.b #SCALE_DATA, &P1IN
  jnz read_scale_next_bit_not_zero
  bis.b #1, r15
read_scale_next_bit_not_zero:
  dec.b r14
  jnz read_scale_next_bit
  mov.b r15, 0(r12)
  inc.w r12
  dec.b r13
  jnz read_scale_next_byte
  ;; Clock an extra bit so it's channel A, Gain 128
  bis.b #SCALE_CLOCK, &P1OUT
  bic.b #SCALE_CLOCK, &P1OUT
  bic.b #LED_0, &P1OUT
  ret

translate_weight:
  mov.w &SCALE_VALUE, r15
  mov.w r15, &SCALE_LAST
  swpb r15
  ;; Subtract the value of no weight on the scale from the value of
  ;; the weight on the scale.
  ;mov.w #0xff39, r14
  mov.w &SCALE_VALUE_0, r14
  sub.w r15, r14
  mov.w r14, r15
  mov.w r15, &SCALE_DIFF
  call #divide
  mov.w r15, &SCALE_FRACTIONAL
  mov.w r14, &SCALE_DECIMAL
  ret

speak_weight:
  call #delay_1s
  ;; Pause.
  mov.b #33, r14
  call #play_file
  call #wait_for_mp3_player
  call #delay_3s
  call #read_scale
  call #translate_weight
  ;; Your weight is.
  mov.b #34, r14
  call #play_file
  call #wait_for_mp3_player
  mov.w &SCALE_DECIMAL, r15
  ;; Find if weight is more than 100 and speak that.
  call #divide_by_100
  mov.w r15, &REMAINDER
  cmp.w #0, r14
  jz speak_weight_less_than_100
  sub.b #29, r14
  call #play_file
  call #wait_for_mp3_player
speak_weight_less_than_100:
  mov.w &REMAINDER, r15
  cmp.w #21, r15
  jlo speak_weight_less_than_20
  ;; Divide current value (between 20 and 99) by 10 to play 2 sound
  ;; files, first for the tens place and last for the ones.
  call #divide_by_10
  mov.w r15, &REMAINDER
  add.w #19, r14
  call #play_file
  call #wait_for_mp3_player
  mov.w &REMAINDER, r14
  add.w #1, r14
  call #play_file
  call #wait_for_mp3_player
  jmp speak_weight_fraction
speak_weight_less_than_20:
  ;; For weights less than 20, just play the sound file.
  mov.w r15, r14
  add.w #1, r14
  call #play_file
  call #wait_for_mp3_player
speak_weight_fraction:
  ;; Say point.
  mov.w #35, r14
  call #play_file
  call #wait_for_mp3_player
  ;; Close enough.
  mov.w &SCALE_FRACTIONAL, r14
  rra.w r14
  rra.w r14
  rra.w r14
  add.w #1, r14
  call #play_file
  call #wait_for_mp3_player
  ;; Say kilograms.
  mov.w #36, r14
  call #play_file
  call #wait_for_mp3_player
  call #delay_3s
  ;; Wait till user steps off scale.
  mov.w #0, r6
speak_weight_wait_step_off:
  call #read_scale
  mov.w &SCALE_VALUE, r15
  swpb r15
  cmp.w #0xff00, r15
  jhs speak_weight_exit
  cmp.w #1, r6
  jz speak_weight_wait_step_off
  mov.w #1, r6
  mov.w #37, r14
  call #play_file
  call #wait_for_mp3_player
  jmp speak_weight_wait_step_off
speak_weight_exit:
  ret

;; get_id()
get_id:
  call #i2c_start
  mov.b #0x6e, r15
  call #i2c_write
  mov.b #0x10, r15
  call #i2c_write
  call #i2c_stop
  call #delay_50ms
  call #i2c_start
  mov.b #0x6f, r15
  call #i2c_write
  call #i2c_read
  mov.b r15, &DEVICE_ID
  call #i2c_stop
  ret

;; set_volume(r14=volume_level)
set_volume:
  call #i2c_start
  mov.b #0x6e, r15
  call #i2c_write
  mov.b #0x07, r15
  call #i2c_write
  mov.b r14, r15
  call #i2c_write
  call #i2c_stop
  ret

;; play_file(r14=song_number)
play_file:
  call #clear_interrupts
  call #i2c_start
  mov.b #0x6e, r15
  call #i2c_write
  mov.b #0x01, r15
  call #i2c_write
  mov.b r14, r15
  call #i2c_write
  call #i2c_stop
  ret

;; clear_interrupts()
clear_interrupts:
  call #i2c_start
  mov.b #0x6e, r15
  call #i2c_write
  mov.b #0x0d, r15
  call #i2c_write
  call #i2c_stop
  ret

;; is_playing() : r15=true/false
is_playing:
  mov.b #0, r15
  bit.b #MP3_INT, &P1IN
  jz is_playing_false
  mov.b #1, r15
is_playing_false:
  ret

wait_for_mp3_player:
  call #is_playing
  cmp.b #1, r15
  jz wait_for_mp3_player
  ret

;; SDA goes low while SCL is still high then SCL goes low.
i2c_start:
  mov.b #0, &USISRL
  bis.b #USIOE|USIGE, &USICTL0
  bic.b #USIGE, &USICTL0
  ret

;; SCL needs to go high and then SDA goes high.
i2c_stop:
  bis.b #USISCLREL, &USICNT
  bis.b #USIOE, &USICTL0
  mov.b #0, &USISRL
  mov.b #1, &USICNT
i2c_stop_busy:
  bit.b #USIIFG, &USICTL1
  jz i2c_stop_busy
  mov.b #0xff, &USISRL
  bis.b #USIGE, &USICTL0
  bic.b #USIOE|USIGE, &USICTL0
  ret

;; i2c_write(data=r15)
i2c_write:
  bis.b #USIOE, &USICTL0
  mov.b r15, &USISRL
  mov.b #8, &USICNT
i2c_write_busy:
  bit.b #USIIFG, &USICTL1
  jz i2c_write_busy
  mov.b #0, &USISRL
  mov.b #1, &USICNT
  bic.b #USIOE, &USICTL0
i2c_write_ack_busy:
  bit.b #USIIFG, &USICTL1
  jz i2c_write_ack_busy
  mov.b &USISRL, r15
  ret

;; i2c_read()
i2c_read:
  bic.b #USIOE, &USICTL0
  ;mov.b r15, &USISRL
  mov.b #8, &USICNT
i2c_read_busy:
  bit.b #USIIFG, &USICTL1
  jz i2c_read_busy
  mov.b &USISRL, r15
  mov.b #0, &USISRL
  mov.b #1, &USICNT
  bis.b #USIOE, &USICTL0
i2c_read_ack_busy:
  bit.b #USIIFG, &USICTL1
  jz i2c_read_ack_busy
  ret

;; divide(r15) -> r14:15
divide:
  mov.w #0, r14
divide_loop:
  cmp.w #84, r15
  jlo divide_loop_exit
  sub.w #84, r15
  add.w #1, r14
  jmp divide_loop
divide_loop_exit:
  ret

;; divide_by_100(r15) -> r14:15
divide_by_100:
  mov.w #0, r14
divide_by_100_loop:
  cmp.w #100, r15
  jlo divide_by_100_loop_exit
  sub.w #100, r15
  add.w #1, r14
  jmp divide_by_100_loop
divide_by_100_loop_exit:
  ret

;; divide_by_10(r15) -> r14:15
divide_by_10:
  mov.w #0, r14
divide_by_10_loop:
  cmp.w #10, r15
  jlo divide_by_10_loop_exit
  sub.w #10, r15
  add.w #1, r14
  jmp divide_by_10_loop
divide_by_10_loop_exit:
  ret

delay_50ms:
  mov.w #467 * 50, r15
delay_loop_1ms:
  dec.w r15
  jnz delay_loop_1ms
  ret

delay_1s:
  mov.w #20, r5
delay_1s_loop:
  call #delay_50ms
  dec.w r5
  jnz delay_1s_loop
  ret

delay_3s:
  mov.w #60, r5
delay_3s_loop:
  call #delay_50ms
  dec.w r5
  jnz delay_3s_loop
  ret

delay_10s:
  mov.w #200, r5
delay_10s_loop:
  call #delay_50ms
  dec.w r5
  jnz delay_10s_loop
  ret

timer_interrupt:
  inc.w r4
  reti

.org 0xfff2
  dw timer_interrupt       ; Timer_A2 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

